/* PSPP - computes sample statistics.
   Copyright (C) 1997, 1998 Free Software Foundation, Inc.
   Written by Ben Pfaff <blp@gnu.org>.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA. */

#include <config.h>
#include <assert.h>
#include <stdlib.h>
#include "common.h"
#include "error.h"
#include "approx.h"
#include "misc.h"
#include "str.h"
#include "lexer.h"
#include "var.h"

/* Implementation details:

   The S?SS manuals do not specify the order that COUNT subcommands are
   performed in.  Experiments, however, have shown that they are performed
   in the order that they are specified in, rather than simultaneously.
   So, with the two variables A and B, and the two cases,

   A B
   1 2
   2 1

   the command COUNT A=A B (1) / B=A B (2) will produce the following
   results,

   A B
   1 1
   1 0

   rather than the results that would be produced if subcommands were
   simultaneous:

   A B
   1 1
   1 1

   Perhaps simultaneity could be implemented as an option.  On the other
   hand, of what use are the above commands?
 */

#undef DEBUGGING
/*#define DEBUGGING 1*/
#include "debug-print.h"

/* Definitions. */

enum
  {
    CNT_ERROR,			/* invalid value */
    CNT_SINGLE,			/* single value */
    CNT_HIGH,			/* x >= a */
    CNT_LOW,			/* x <= a */
    CNT_RANGE,			/* a <= x <= b */
    CNT_ANY,			/* count any */
    CNT_SENTINEL		/* list terminator */
  };

typedef struct
  {
    int type;
    double a, b;
  }
cnt_num;

typedef struct
  {
    int type;
    char *s;
  }
cnt_str;

typedef struct counting counting;
struct counting
  {
    counting *next;

    /* variables to count */
    variable **v;
    int n;

    /* values to count */
    int missing;		/* (numeric only) 0=don't count missing,
				   1=count SYSMIS,
				   2=count system- and user-missing */
    union			/* Criterion values. */
      {
	cnt_num *n;
	cnt_str *s;
      }
    crit;
  }
/* counting */ ;

typedef struct cnt_var_info cnt_var_info;
struct cnt_var_info
  {
    cnt_var_info *next;

    variable *d;		/* Destination variable. */
    char n[9];			/* Name of dest var. */

    counting *c;		/* The counting specifications. */
  }
/* cnt_var_info */ ;

typedef struct
  {
    trns_header h;
    cnt_var_info *specs;
  }
count_trns;

#if DEBUGGING
static void debug_print (void);
#endif

/* First counting in chain. */
static cnt_var_info *head;

/* Parser. */

static int count_trns_proc (any_trns *, ccase *);
static void count_trns_free (any_trns *);

static int parse_numeric_criteria (counting *);
static int parse_string_criteria (counting *);

int cmd_count (void);

int
internal_cmd_count (void)
{
  int code = cmd_count ();
  if (!code)
    {
      count_trns c;
      c.specs = head;
      count_trns_free ((any_trns *) & c);
    }
  return code;
}

int
cmd_count (void)
{
  /* Specification currently being parsed. */
  cnt_var_info *cnt;

  /* Counting currently being parsed. */
  counting *c;

  /* Return value from parsing function. */
  int ret;

  /* Transformation. */
  count_trns *trns;

  match_id (COUNT);

  /* Parses each slash-delimited specification. */
  head = cnt = xmalloc (sizeof (cnt_var_info));
  while (1)
    {
      /* Initialize this cnt_var_info to ensure proper cleanup. */
      cnt->next = NULL;
      cnt->d = NULL;
      cnt->c = NULL;

      /* Get destination variable, or at least its name. */
      force_id ();
      cnt->d = find_variable (tokstr);
      if (cnt->d)
	{
	  if (cnt->d->type == ALPHA)
	    {
	      msg (SE, _("Destination cannot be a string variable."));
	      goto fail;
	    }
	}
      else
	strcpy (cnt->n, tokstr);

      get_token ();
      force_match ('=');

      c = cnt->c = xmalloc (sizeof (counting));
      while (1)
	{
	  c->next = NULL;
	  c->v = NULL;
	  if (!parse_variables (NULL, &c->v, &c->n, PV_DUPLICATE | PV_SAME_TYPE))
	    goto fail;

	  force_match ('(');

	  ret = (c->v[0]->type == NUMERIC
		 ? parse_numeric_criteria
		 : parse_string_criteria) (c);
	  if (!ret)
	    goto fail;

	  if (token == '/' || token == '.')
	    break;

	  c = c->next = xmalloc (sizeof (counting));
	}

      if (token == '.')
	break;

      force_match ('/');
      cnt = cnt->next = xmalloc (sizeof (cnt_var_info));
    }

  /* Create all the nonexistent destination variables. */
  for (cnt = head; cnt; cnt = cnt->next)
    if (!cnt->d)
      {
	/* It's legal, though motivationally questionable, to count to
	   the same dest var more than once. */
	cnt->d = find_variable (cnt->n);

	if (!cnt->d)
	  cnt->d = force_create_variable (&default_dict, cnt->n, NUMERIC, 0);
      }

#if DEBUGGING
  debug_print ();
#endif

  trns = xmalloc (sizeof (count_trns));
  trns->h.proc = count_trns_proc;
  trns->h.free = count_trns_free;
  trns->specs = head;
  add_transformation ((any_trns *) trns);

  return 1;

fail:
  {
    count_trns t;
    t.specs = head;
    count_trns_free ((any_trns *) & t);
    return 0;
  }
}

/* Parses a set of numeric criterion values. */
static int
parse_numeric_criteria (counting * c)
{
  int n = 0;
  int m = 0;

  c->crit.n = 0;
  c->missing = 0;
  while (1)
    {
      cnt_num *cur;
      if (n >= m - 1)
	{
	  m += 16;
	  c->crit.n = xrealloc (c->crit.n, m * sizeof (cnt_num));
	}

      cur = &c->crit.n[n++];
      if (token == NUM)
	{
	  cur->a = tokval;
	  get_token ();
	  if (match_id (THRU))
	    {
	      if (token == NUM)
		{
		  force_num ();
		  cur->b = tokval;
		  cur->type = CNT_RANGE;
		  get_token ();

		  if (cur->a > cur->b)
		    return msg (SE, _("%g THRU %g is not a valid range.  The "
				"number following THRU must be at least "
				"as big as the number preceding THRU."),
				cur->a, cur->b);
		}
	      else if (match_id (HI) || match_id (HIGHEST))
		cur->type = CNT_HIGH;
	      else
		return syntax_error (NULL);
	    }
	  else
	    cur->type = CNT_SINGLE;
	}
      else if (match_id (LO) || match_id (LOWEST))
	{
	  force_match_id (THRU);
	  if (token == NUM)
	    {
	      cur->type = CNT_LOW;
	      cur->a = tokval;
	      get_token ();
	    }
	  else if (match_id (HI) || match_id (HIGHEST))
	    cur->type = CNT_ANY;
	  else
	    return syntax_error (NULL);
	}
      else if (match_id (SYSMIS))
	{
	  if (c->missing < 1)
	    c->missing = 1;
	}
      else if (match_id (MISSING))
	c->missing = 2;
      else
	return syntax_error (NULL);

      match_tok (',');
      if (match_tok (')'))
	break;
    }

  c->crit.n[n].type = CNT_SENTINEL;
  return 1;
}

/* Parses a set of string criteria values.  The skeleton is the same
   as parse_numeric_criteria(). */
static int
parse_string_criteria (counting * c)
{
  int len = 0;

  int n = 0;
  int m = 0;

  int i;

  for (i = 0; i < c->n; i++)
    if (c->v[i]->width > len)
      len = c->v[i]->width;

  c->crit.n = 0;
  while (1)
    {
      cnt_str *cur;
      if (n >= m - 1)
	{
	  m += 16;
	  c->crit.n = xrealloc (c->crit.n, m * sizeof (cnt_str));
	}

      force_string ();
      cur = &c->crit.s[n++];
      cur->type = CNT_SINGLE;
      cur->s = malloc (len + 1);
      strpadcpy (cur->s, tokstr, len + 1, ' ');
      get_token ();

      match_tok (',');
      if (match_tok (')'))
	break;
    }

  c->crit.s[n].type = CNT_SENTINEL;
  return 1;
}

/* Transformation. */

/* Counts the number of values in case C matching counting CNT. */
static inline int
count_numeric (counting * cnt, ccase * c)
{
  int counter = 0;

  cnt_num *num;

  double cmp;
  int i;

  for (i = 0; i < cnt->n; i++)
    {
      /* Extract the variable value and eliminate missing values. */
      cmp = c->data[cnt->v[i]->fv].f;
      if (cmp == SYSMIS)
	{
	  if (cnt->missing >= 1)
	    counter++;
	  continue;
	}
      if (cnt->missing >= 2 && is_num_user_missing (cmp, cnt->v[i]))
	{
	  counter++;
	  continue;
	}

      /* Try to find the value in the list. */
      for (num = cnt->crit.n;; num++)
	switch (num->type)
	  {
	  case CNT_ERROR:
	    assert (0);
	    break;
	  case CNT_SINGLE:
	    if (approx_ne (cmp, num->a))
	      break;
	    counter++;
	    goto done;
	  case CNT_HIGH:
	    if (approx_lt (cmp, num->a))
	      break;
	    counter++;
	    goto done;
	  case CNT_LOW:
	    if (approx_gt (cmp, num->a))
	      break;
	    counter++;
	    goto done;
	  case CNT_RANGE:
	    if (approx_lt (cmp, num->a) || approx_gt (cmp, num->b))
	      break;
	    counter++;
	    goto done;
	  case CNT_ANY:
	    counter++;
	    goto done;
	  case CNT_SENTINEL:
	    goto done;
	  default:
	    assert (0);
	  }
    done: ;
    }
  return counter;
}

/* Counts the number of values in case C matching counting CNT. */
static inline int
count_string (counting * cnt, ccase * c)
{
  int counter = 0;

  cnt_str *str;

  char *cmp;
  int len;

  int i;

  for (i = 0; i < cnt->n; i++)
    {
      /* Extract the variable value, variable width. */
      cmp = c->data[cnt->v[i]->fv].s;
      len = cnt->v[i]->width;

      for (str = cnt->crit.s;; str++)
	switch (str->type)
	  {
	  case CNT_ERROR:
	    assert (0);
	  case CNT_SINGLE:
	    if (memcmp (cmp, str->s, len))
	      break;
	    counter++;
	    goto done;
	  case CNT_SENTINEL:
	    goto done;
	  default:
	    assert (0);
	  }
    done: ;
    }
  return counter;
}

/* Performs the COUNT transformation T on case C. */
static int
count_trns_proc (any_trns * trns, ccase * c)
{
  cnt_var_info *info;
  counting *cnt;
  int counter;

  for (info = ((count_trns *) trns)->specs; info; info = info->next)
    {
      counter = 0;
      for (cnt = info->c; cnt; cnt = cnt->next)
	if (cnt->v[0]->type == NUMERIC)
	  counter += count_numeric (cnt, c);
	else
	  counter += count_string (cnt, c);
      c->data[info->d->fv].f = counter;
    }
  return -1;
}

/* Destroys all dynamic data structures associated with T. */
static void
count_trns_free (any_trns * t)
{
  cnt_var_info *iter, *next;

  for (iter = ((count_trns *) t)->specs; iter; iter = next)
    {
      counting *i, *n;

      for (i = iter->c; i; i = n)
	{
	  if (i->n && i->v)
	    {
	      if (i->v[0]->type == NUMERIC)
		free (i->crit.n);
	      else
		{
		  cnt_str *s;

		  for (s = i->crit.s; s->type != CNT_SENTINEL; s++)
		    free (s->s);
		  free (i->crit.s);
		}
	    }
	  free (i->v);

	  n = i->next;
	  free (i);
	}

      next = iter->next;
      free (iter);
    }
}

/* Debugging. */

#if DEBUGGING
static void
debug_print (void)
{
  cnt_var_info *iter;
  counting *i;
  int j;

  printf ("COUNT\n");
  for (iter = head; iter; iter = iter->next)
    {
      printf ("  %s=", iter->d->name);
      for (i = iter->c; i; i = i->next)
	{
	  for (j = 0; j < i->n; j++)
	    printf ("%s%s", j ? " " : "", i->v[j]->name);
	  printf (" (");
	  if (i->v[0]->type == NUMERIC)
	    {
	      cnt_num *n;

	      if (i->missing == 2)
		printf ("MISSING");
	      else if (i->missing == 1)
		printf ("SYSMIS");
	      else
		assert (i->missing == 0);

	      for (n = i->crit.n; n->type != CNT_SENTINEL; n++)
		{
		  if (i->missing && n != i->crit.n)
		    printf (",");
		  switch (n->type)
		    {
		    case CNT_SINGLE:
		      printf ("%g", n->a);
		      break;
		    case CNT_HIGH:
		      printf ("%g THRU HIGH", n->a);
		      break;
		    case CNT_LOW:
		      printf ("LOW THRU %g", n->a);
		      break;
		    case CNT_RANGE:
		      printf ("%g THRU %g", n->a, n->b);
		      break;
		    case CNT_ANY:
		      printf ("LOW THRU HIGH");
		      break;
		    default:
		      printf ("<ERROR %d>", n->type);
		      break;
		    }
		}
	    }
	  else
	    {
	      cnt_str *s;

	      for (s = i->crit.s; s->type != CNT_SENTINEL; s++)
		{
		  if (s != i->crit.s)
		    printf (",");
		  if (s->type == CNT_SINGLE)
		    printf ("'%s'", s->s);
		  else
		    printf ("<ERROR %d>", s->type);
		}
	    }
	  printf (")  ");
	}
      printf ("\n");
    }
}
#endif /* DEBUGGING */
