/* PSPP - computes sample statistics.
   Copyright (C) 1997, 1998 Free Software Foundation, Inc.
   Written by Ben Pfaff <blp@gnu.org>.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA. */

#include <config.h>

#if TIME_WITH_SYS_TIME
#include <sys/time.h>
#include <time.h>
#else
#if HAVE_SYS_TIME_H
#include <sys/time.h>
#else
#include <time.h>
#endif
#endif

#include <assert.h>
#include <ctype.h>
#include <math.h>
#include <stdlib.h>
#include "common.h"
#include "error.h"
#include "str.h"
#include "misc.h"
#include "settings.h"
#include "lexer.h"
#include "var.h"

/* Returns ceil(log10(X)), x>0; 1, x==0; undefined, x<0.  Basically
   returns the number of digits in X for nonnegative X. */
int
intlog10 (int x)
{
  static const long tab[] =
  {10, 100, 1000, 10000, 100000, 1000000, 10000000, 10000000,
   1000000000, 0};
  const long *tp;

  int i;

  assert (x >= 0);
  for (i = 1, tp = tab; *tp; i++, tp++)
    if (x < *tp)
      return i;
  assert (0);
#if __GNUC__ || __BORLANDC__
  return 0;
#endif
}

/* Returns a null-terminated string of N spaces. */
char *
spacing (int n)
{
  static char *s;
  static int max = 0;

  if (n + 1 > max)
    {
      if (s)
	free (s);
      s = xmalloc (n + 1);
      max = n + 1;
    }
  memset (s, ' ', n);
  s[n] = 0;
  return s;
}

/* Some systems are so broken that they do not supply a value for
   RAND_MAX.  There is absolutely no reliable way to determine this
   value, either.  So we must supply our own.  This one is the one
   presented in the ANSI C standard as strictly compliant. */
#if BROKEN_RAND
static unsigned long int next = 1;

int
ansi_rand (void)
{
  next = next * 1103515245 + 12345;
  return (unsigned int)(next / 65536) % 32768;
}

void
ansi_srand (unsigned int seed)
{
  next = seed;
}
#endif /* BROKEN_RAND */

/* The random number generator here is an implementation in C of
   Knuth's Algorithm 3.2.2B (Randomizing by Shuffling) in _The Art of
   Computer Programming_, Vol. 2. */

#define k 13
static int V[k];
static int Y;

static double X2;

/* Initializes the random number generator.  Should be called once by
   every cmd_*() that uses random numbers.  Note that this includes
   all procedures that use expressions since they may generate random
   numbers. */
void
setup_randomize (void)
{
  static time_t curtime;
  int i;

  if (set_seed == NOT_LONG)
    {
      if (!curtime)
	time (&curtime);
      srand (curtime++);
    }
  else
    srand (set_seed);

  set_seed_used = 1;

  for (i = 0; i < k; i++)
    V[i] = rand ();
  Y = rand ();
  X2 = SYSMIS;
}

/* Standard shuffling procedure for increasing randomness of the ANSI
   C random number generator. Returns a random number R where 0 <= R
   <= RAND_MAX. */
inline int
shuffle (void)
{
  int j = k * Y / RAND_MAX;
  Y = V[j];
  V[j] = rand ();
  return Y;
}

/* Returns a random number R where 0 <= R <= X. */
double 
rand_uniform (double x)
{
  return ((double) shuffle ()) / (((double) RAND_MAX) / x);
}

/* Returns a random number from the distribution with mean 0 and
   standard deviation X.  This uses algorithm P in section 3.4.1C of
   Knuth's _Art of Computer Programming_, Vol 2. */
double 
rand_normal (double x)
{
  double U1, U2;
  double V1, V2;
  double S;
  double X1;

  if (X2 != SYSMIS)
    {
      double t = X2;
      X2 = SYSMIS;
      return t * x;
    }
  do
    {
      U1 = ((double) shuffle ()) / RAND_MAX;
      U2 = ((double) shuffle ()) / RAND_MAX;
      V1 = 2 * U1 - 1;
      V2 = 2 * U2 - 1;
      S = V1 * V1 + V2 * V2;
    }
  while (S >= 1);
  X1 = V1 * sqrt (-2. * log (S) / S);
  X2 = V2 * sqrt (-2. * log (S) / S);
  return X1 * x;
}

/* Returns a random integer R, where 0 <= R < X. */
int
rand_simple (int x)
{
  return shuffle () % x;
}

/* Reads a line from FILE into *LINE which can hold *N bytes, in the
   fashion of getline().  Splices lines terminated with `\', then
   deletes comments introduced by `#' outside of single or double
   quotes.  Trailing whitespace will be deleted.  Increments cust_ln
   as appropriate. */
int
get_config_line (char **line, size_t * n, FILE * file)
{
  char *cp;
  int len = getline (line, n, file);

  cust_ln++;
  if (-1 == len)
    return -1;

  while (len && isspace ((unsigned char) (*line)[len - 1]))
    len--;
  (*line)[len] = 0;

  if (len == 0 || '\\' != (*line)[len - 1])
    goto exit;
  (*line)[--len] = 0;
  while (1)
    {
      char *line2 = NULL;
      size_t n2 = 0;
      int len2 = getline (&line2, &n2, file);

      cust_ln++;
      if (-1 == len2)
	return len2;

      while (len2 && isspace ((unsigned char) line2[len2 - 1]))
	len2--;
      line2[len2] = 0;

      *line = xrealloc (*line, len + len2 + 1);
      *n = len + len2 + 1;
      strcpy (&(*line)[len], line2);
      len += len2;
      free (line2);

      if (len == 0 || '\\' != (*line)[len - 1])
	goto exit;
      (*line)[--len] = 0;
    }

exit:
  cp = strchr (*line, '#');
  if (cp)
    {
      int quote = 0;
      for (cp = *line; *cp; cp++)
	if (quote)
	  {
	    if (*cp == quote)
	      quote = 0;
	    else if (*cp == '\\')
	      cp++;
	  }
	else
	  {
	    if (*cp == '\'' || *cp == '"')
	      quote = *cp;
	    else if (*cp == '#')
	      {
		len = cp - *line;
		while (len && isspace ((unsigned char) (*line)[len]))
		  len--;
		(*line)[len] = 0;
		break;
	      }
	  }
    }
  return len;
}

#if 0
/* This function is not currently used, however it might be useful
   someday.  It reverses the order of the COUNT elements of ARRAY,
   each of which is SIZE bytes in size. */
static void
reverse (void *_array, int count, int size)
{
  char *array = _array;
  void *temp;
  int i;

  temp = local_alloc (size);

  for (i = 0; i < count / 2; i++)
    {
      memcpy (temp, &array[i * size], size);
      memcpy (&array[i * size], &array[(count - i) * size], size);
      memcpy (&array[(count - i) * size], temp, size);
    }
  local_free (temp);
}
#endif
